---@class NixieStorage
---@field alphas {[integer]:LuaEntity?}
---@field next_alpha? integer
---@field controllers {[integer]:LuaEntity?}
---@field next_controller? integer
---@field nextdigit {[integer]:LuaEntity?}
---@field cache {[integer]:NixieCache?}
storage = {}

---@class NixieCache
---@field control? LuaLampControlBehavior
---@field flags? integer
---@field lastvalue? integer
---@field lastcolor Color[]
---@field sprites LuaRenderObject[]


---@param unit_number integer
---@return NixieCache
local function getCache(unit_number)
  local cache = storage.cache[unit_number]
  if not cache then
    cache = {
      lastcolor = {},
      sprites = {},
    }
    storage.cache[unit_number] = cache
  end
  return cache
end


local validEntityName = {
  ['nixie-tube']       = 1,
  ['nixie-tube-alpha'] = 1,
  ['nixie-tube-small'] = 2
}

local signalCharMap = {
  ["signal-0"] = "0",
  ["signal-1"] = "1",
  ["signal-2"] = "2",
  ["signal-3"] = "3",
  ["signal-4"] = "4",
  ["signal-5"] = "5",
  ["signal-6"] = "6",
  ["signal-7"] = "7",
  ["signal-8"] = "8",
  ["signal-9"] = "9",
  ["signal-A"] = "A",
  ["signal-B"] = "B",
  ["signal-C"] = "C",
  ["signal-D"] = "D",
  ["signal-E"] = "E",
  ["signal-F"] = "F",
  ["signal-G"] = "G",
  ["signal-H"] = "H",
  ["signal-I"] = "I",
  ["signal-J"] = "J",
  ["signal-K"] = "K",
  ["signal-L"] = "L",
  ["signal-M"] = "M",
  ["signal-N"] = "N",
  ["signal-O"] = "O",
  ["signal-P"] = "P",
  ["signal-Q"] = "Q",
  ["signal-R"] = "R",
  ["signal-S"] = "S",
  ["signal-T"] = "T",
  ["signal-U"] = "U",
  ["signal-V"] = "V",
  ["signal-W"] = "W",
  ["signal-X"] = "X",
  ["signal-Y"] = "Y",
  ["signal-Z"] = "Z",
  ["signal-negative"] = "negative",

  --extended symbols
  ["signal-stop"] = "dot",
  ["signal-qmark"]="?",
  ["signal-exmark"]="!",
  ["signal-at"]="@",
  ["signal-sqopen"]="[",
  ["signal-sqclose"]="]",
  ["signal-curopen"]="{",
  ["signal-curclose"]="}",
  ["signal-paropen"]="(",
  ["signal-parclose"]=")",
  ["signal-slash"]="slash",
  ["signal-asterisk"]="*",
  ["signal-minus"]="-",
  ["signal-plus"]="+",
  ["signal-percent"]="%",
}

--sets the state(s) and update the sprite for a nixie
local is_simulation = script.level.is_simulation

---@param nixie LuaEntity
---@param cache NixieCache
---@param newstates string[]
---@param newcolor? Color
local function setStates(nixie,cache,newstates,newcolor)
  for key,new_state in pairs(newstates) do
    if not new_state then new_state = "off" end
    -- printing floats sometimes hands us a literal '.', needs to be renamed
    if new_state == '.' then new_state = "dot" end


    local obj = cache.sprites[key]
    if not (obj and obj.valid) then
      cache.lastcolor[key] = nil

      local num = validEntityName[nixie.name]
      ---@type Vector.0
      local position
      if num == 1 then -- large tube, one sprite
        position = {x=1/32, y=1/32}
      else
        position = {x=-9/64+((key-1)*20/64), y=3/64} -- sprite offset
      end
      obj = rendering.draw_sprite{
        sprite = "nixie-tube-sprite-" .. new_state,
        target = { entity = nixie, offset = position},
        surface = nixie.surface,
        tint = {r=1.0,  g=1.0,  b=1.0, a=1.0},
        x_scale = 1/num,
        y_scale = 1/num,
        render_layer = "object",
        }

      cache.sprites[key] = obj
    end

    if nixie.energy > 70 or is_simulation then
      obj.sprite = "nixie-tube-sprite-" .. new_state

      local color = newcolor
      if not color then color = {r=1.0,  g=0.6,  b=0.2, a=1.0} end
      if new_state == "off" then color={r=1.0,  g=1.0,  b=1.0, a=1.0} end

      if not (cache.lastcolor[key] and (cache.lastcolor[key].r == color.r) and (cache.lastcolor[key].g == color.g) and (cache.lastcolor[key].b == color.b) and (cache.lastcolor[key].a == color.a)) then
        cache.lastcolor[key] = color
        obj.color = color
      end
    else
      if obj.sprite ~= "nixie-tube-sprite-off" then
        obj.sprite = "nixie-tube-sprite-off"
      end
      obj.color = {r=1.0,  g=1.0,  b=1.0, a=1.0}
      cache.lastcolor[key] = nil
    end
  end
end

---@param behavior LuaLampControlBehavior
---@return SignalID?
local function get_selected_signal(behavior)
  if behavior == nil then
    return nil
  end

	local condition = behavior.circuit_condition
	if condition == nil then
    return nil
  end

  local signal = condition.first_signal
  if signal and not condition.fulfilled then
    -- use >= MININT32 to ensure always-on
    condition.comparator="≥"
    condition.constant=-0x80000000
    condition.second_signal=nil
    behavior.circuit_condition = condition
  end

  return signal
end

---@param entity LuaEntity
---@param vs? string
---@param color? Color
local function displayValString(entity,vs,color)
  local offset = vs and #vs or 0
  while entity do
    ---@type LuaEntity?
    local nextdigit = storage.nextdigit[entity.unit_number]
    local cache = getCache(entity.unit_number)
    local chcount = #cache.sprites

    if not vs then
      setStates(entity,cache,(chcount==1) and {"off"} or {"off","off"})
    elseif offset < chcount then
      setStates(entity,cache,{"off",vs:sub(offset,offset)},color)
    elseif offset >= chcount then
      setStates(entity,cache,
        (chcount==1) and
          {vs:sub(offset,offset)} or
          {vs:sub(offset-1,offset-1),vs:sub(offset,offset)}
        ,color)
    end

    if nextdigit then
      if nextdigit.valid then
        if offset>chcount then
          offset = offset-chcount
        else
          vs = nil
        end
      else
        --when a nixie in the middle is removed, it doesn't have the unit_number to it's right to remove itself
        storage.nextdigit[entity.unit_number] = nil
        nextdigit = nil
      end
    end
    ---@diagnostic disable-next-line:cast-local-type
    entity = nextdigit
  end
end

---@param i integer
---@return float
local function float_from_int(i)
  local sign = bit32.btest(i,0x80000000) and -1 or 1
  local exponent = bit32.rshift(bit32.band(i,0x7F800000),23)-127
  local significand = bit32.band(i,0x007FFFFF)

  if exponent == 128 then
    if significand == 0 then
      return sign/0 --[[infinity]]
    else
      return 0/0 --[[nan]]
    end
  end

  if exponent == -127 then
    if significand == 0 then
      return sign * 0 --[[zero]]
    else
      return sign * math.ldexp(significand,-149) --[[denormal numbers]]
    end
  end

  return sign * math.ldexp(bit32.bor(significand,0x00800000),exponent-23) --[[normal numbers]]
end

---@param high integer
---@param low integer
---@return double
local function double_from_ints(high, low)
  local sign = bit32.btest(high,0x80000000) and -1 or 1
  local exponent = bit32.rshift(bit32.band(high,0x7FF00000),20)-1023
  -- have to use regular math to reassemble this becuase it's so long
  local significand = (bit32.band(high,0x000FFFFF) * 0x100000000) + low

  if exponent == 1024 then
    if significand == 0 then
      return sign/0 --[[infinity]]
    else
      return 0/0 --[[nan]]
    end
  end

  if exponent == -1023 then
    if significand == 0 then
      return sign * 0 --[[zero]]
    else
      return sign * math.ldexp(significand,-1074) --[[denormal numbers]]
    end
  end

  return sign * math.ldexp(significand+0x10000000000000,exponent-52) --[[normal numbers]]
end

---@param entity LuaEntity
---@return string?
local function getAlphaSignals(entity)
  local signals = entity.get_signals(defines.wire_connector_id.circuit_red, defines.wire_connector_id.circuit_green)
  ---@type string
  local ch

  if signals and #signals > 0 then
    for _,s in pairs(signals) do
      if signalCharMap[s.signal.name] then
        if ch then
          return "err"
        else
          ch = signalCharMap[s.signal.name]
        end
      end
    end
  end

  return ch
end

---@type SignalID
local sigFloat = {name="signal-float",type="virtual"}

---@type SignalID
local sigHex = {name="signal-hex",type="virtual"}

---@param entity LuaEntity
---@param cache NixieCache
local function onTickController(entity,cache)
  local control = cache.control
  if not (control and control.valid) then
    control = entity.get_or_create_control_behavior() --[[@as LuaLampControlBehavior]]
    cache.control = control
  end

  local float_v = entity.get_signal(sigFloat, defines.wire_connector_id.circuit_red, defines.wire_connector_id.circuit_green)
  local float = float_v == 1
  local double = float_v == 2
  
  local hex = entity.get_signal(sigHex, defines.wire_connector_id.circuit_green, defines.wire_connector_id.circuit_red) ~= 0

  local selected = get_selected_signal(control)
  ---@type number, number
  local v = 0
  if selected then
    if double then
      v = double_from_ints(
        entity.get_signal(selected, defines.wire_connector_id.circuit_green),
        entity.get_signal(selected, defines.wire_connector_id.circuit_red))
    else
      v = entity.get_signal(selected, defines.wire_connector_id.circuit_red, defines.wire_connector_id.circuit_green)
      if float then
        v = float_from_int(v)
      end
    end
  end

  local use_colors = cache.control.use_colors
  local flags = (float_v) + (hex and 4 or 0) + (use_colors and 8 or 0)

  --if value or any flags changed, or always update while use_colors...
  if cache.lastvalue ~= v or use_colors or flags ~= cache.flags then
    cache.flags = flags
    cache.lastvalue = v

    
    local format = "%i"
    if float then
      format = "%G"
      if hex then
        format = "%A"
      end
    elseif double then
      format = "%.14G"
      if hex then
        format = "%A"
      end
    else
      if hex then
        format = "%X"
        if v < 0 then v = v + 0x100000000 end
      end
    end

    displayValString(entity,format:format(v),control.use_colors and control.color or nil)
  end

end

local always_on = {
  condition={
    first_signal={name="signal-anything",type="virtual"},
    comparator="≠",
    constant=0,
    second_signal=nil
  },
  connect_to_logistic_network=false
}

---@param entity LuaEntity
---@param cache NixieCache
local function onTickAlpha(entity,cache)
  local charsig = getAlphaSignals(entity) or "off"

  ---@type Color?
  local color
  local control = cache.control
  if not (control and control.valid) then
    control = entity.get_or_create_control_behavior() --[[@as LuaLampControlBehavior]]
    cache.control = control
  end
  if control.use_colors then
    control.circuit_condition = always_on
    color = control.color
  end

  setStates(entity,cache,{charsig},color)
end

script.on_event(defines.events.on_tick, function()
  for _=1, settings.global["nixie-tube-update-speed-numeric"].value do
    ---@type LuaEntity?
    local nixie
    if storage.next_controller and not storage.controllers[storage.next_controller] then
      storage.next_controller=nil
    end

    
    storage.next_controller,nixie = next(storage.controllers,storage.next_controller)
    
    ::retry::
    if nixie then
      if nixie.valid then
        onTickController(nixie,getCache(storage.next_controller))
      else
        -- advance now before removal...
        local gone = storage.next_controller
        ---@cast gone -?
        storage.next_controller,nixie = next(storage.controllers,storage.next_controller)

        -- promote next digit if any
        local nextdigit = storage.nextdigit[gone]
        if nextdigit then
          if nextdigit.valid then
            storage.controllers[nextdigit.unit_number] = nextdigit
            displayValString(nextdigit)
          end
          storage.nextdigit[gone] = nil
        end

        storage.controllers[gone] = nil
        storage.cache[gone] = nil
        goto retry
      end
    else
      break
    end
  end

  for _=1, settings.global["nixie-tube-update-speed-alpha"].value do
    ---@type LuaEntity?
    local nixie
    if storage.next_alpha and not storage.alphas[storage.next_alpha] then
      storage.next_alpha=nil
    end
    
    storage.next_alpha,nixie = next(storage.alphas,storage.next_alpha)

    ::retry::
    if nixie then
      if nixie.valid then
        onTickAlpha(nixie, getCache(storage.next_alpha))
      else
        -- advance now before removal...
        local gone = storage.next_alpha
        ---@cast gone -?
        storage.next_alpha,nixie = next(storage.alphas,storage.next_alpha)

        storage.alphas[gone] = nil
        storage.cache[gone] = nil
        goto retry
      end
    else
      break
    end
  end
end)

script.on_event(defines.events.on_object_destroyed, function (event)
  if event.type ~= defines.target_type.entity then return end
  
  -- promote next digit if any
  local nextdigit = storage.nextdigit[event.useful_id]
  if nextdigit then
    if nextdigit.valid then
      storage.controllers[nextdigit.unit_number] = nextdigit
      displayValString(nextdigit)
    end
    storage.nextdigit[event.useful_id] = nil
  end


end)

---@param entity LuaEntity
local function onPlaceEntity(entity)
  local num = validEntityName[entity.name]
  if num then
    script.register_on_object_destroyed(entity)
    local surf=entity.surface
    local cache =getCache(entity.unit_number)
    local sprites = cache.sprites
    for n=1, num do
      --place the /real/ thing(s) at same spot
      ---@type Vector
      local position
      if num == 1 then -- large tube, one sprite
        position = {x=1/32, y=1/32}
      else
        position = {x=-9/64+((n-1)*20/64), y=3/64} -- sprite offset
      end
      local sprite= rendering.draw_sprite{
        sprite = "nixie-tube-sprite-off",
        target = { entity = entity, offset = position},
        surface = entity.surface,
        tint = {r=1.0,  g=1.0,  b=1.0, a=1.0},
        x_scale = 1/num,
        y_scale = 1/num,
        render_layer = "object",
        }

      sprites[n]=sprite
    end

    cache.control = entity.get_or_create_control_behavior() --[[@as LuaLampControlBehavior]]

    if entity.name == "nixie-tube-alpha" then
      storage.alphas[entity.unit_number] = entity
    else

      --enslave guy to left, if there is one
      local neighbors=surf.find_entities_filtered{
        position={x=entity.position.x-1,y=entity.position.y},
        name=entity.name}
      for _,n in pairs(neighbors) do
        if n.valid then
          if storage.next_controller == n.unit_number then
            -- if it's currently the *next* controller, clear that
            storage.next_controller = nil
          end
          storage.controllers[n.unit_number] = nil
          storage.nextdigit[entity.unit_number] = n
        end
      end

      --slave self to right, if any
      neighbors=surf.find_entities_filtered{
        position={x=entity.position.x+1,y=entity.position.y},
        name=entity.name}
      local foundright=false
      for _,n in pairs(neighbors) do
        if n.valid then
          foundright=true
          storage.nextdigit[n.unit_number]=entity
        end
      end
      if not foundright then
        storage.controllers[entity.unit_number] = entity
      end
    end
  end
end


script.on_init(function()
  storage.alphas = {}
  storage.controllers = {}
  storage.cache = {}
  storage.nextdigit = {}
end)

local function RebuildNixies()
  -- clear the tables
  storage = {
    alphas = {},
    controllers = {},
    cache = {},
    nextdigit = {},
  }

  -- wipe out any lingering sprites i've just deleted the references to...
  rendering.clear("nixie-tubes")

  -- and re-index the world
  for _,surf in pairs(game.surfaces) do
    -- re-index all nixies. non-nixie lamps will be ignored by onPlaceEntity
    for _,lamp in pairs(surf.find_entities_filtered{type="lamp"}) do
      onPlaceEntity(lamp)
    end
  end
end

remote.add_interface("nixie-tubes",{
  RebuildNixies = RebuildNixies
})

commands.add_command("RebuildNixies","Reset all Nixie Tubes to clear display glitches.", RebuildNixies)

script.on_configuration_changed(function(data)
  if data.mod_changes and data.mod_changes["nixie-tubes"] then
    RebuildNixies()
  end
end)

local filters = {
}
local names = {}
for name in pairs(validEntityName) do
  filters[#filters+1] = {filter="name",name=name}
  filters[#filters+1] = {filter="ghost_name",name=name}
  names[#names+1] = name
end


script.on_event(defines.events.on_script_trigger_effect, function (event)
  if event.effect_id == "nixie-created" then
    local entity = event.cause_entity
    if entity then
      onPlaceEntity(entity)
    end
  end
end)
